% analyze_recorder.m
% Analyze a recorded sound of a tenor recorder.
% The time offsets of the steady state and the attack transients are marked
% manually. (This can also be done using a trigger signal, as it was the
% case for our JSV paper.)

clear;

% Read the audio sample
% E4 note played on the tenor recorder, 3 times
% Blowing pressure: 120 Pa
[x, fs] = audioread('samples/tenor_E4_0120Pa.wav');

% Create time vector
n = size(x, 1);
t = (0 : n-1).' / fs;

% Create a hi-pass filter for removing low frequency noise
[b, a] = butter(3, 50/(fs/2), 'high');
x = filter(b, a, x);

%% Plot the original signal
figure;
plot(t, x);
xlabel('Time [s]');
ylabel('Signal');

%% Define segments
% Start and end time offsets of the steady and attack transients for the
% three successive sounds
t_steady = [4 8; 12 16; 21 25];
t_attack = [3.725 4.225; 11.295 11.795; 20.350 20.850];
n_segments = size(t_steady, 1);

blks = cell(n_segments, 1);
signals = cell(n_segments, 1);

n_harms = 9;    % Number of harmonics
%% Process each segment
for i_segment = 1 : n_segments
    % Select steady part to estimate fundamental frequency
    sel = t >= t_steady(i_segment, 1) & t < t_steady(i_segment, 2);
    % Estimate fundamental frequency
    f1 = estimate_fund_freq(x(sel), fs);
    fprintf('Segment #%d - Estimated f1 = %.2f\n', i_segment, f1);
    
    % Select attack part for analysis
    sel = t >= t_attack(i_segment, 1) & t < t_attack(i_segment, 2);
    f_n = f1 * (1 : n_harms);
    [blks{i_segment}, signals{i_segment}] = track_f_a(x(sel), fs, f_n);
end

%% Plot reconstruction
i_segment =  1;
sel = t >= t_attack(i_segment, 1) & t < t_attack(i_segment, 2);

figure;
plot(t(sel), x(sel), t(sel), signals{i_segment}.fullfit);
xlabel('Time [s]');
ylabel('Signals');
legend({'Recording', 'Reconstructed'}, 'Location', 'NorthEast');
title(sprintf('Reconstruction of segment #%d', i_segment));
%% Plot frequency dependence
figure;
ax1 = subplot(2, 1, 1); hold on;
ax2 = subplot(2, 1, 2); hold on;
for i_segment = 1 : n_segments
    plot(ax1, blks{i_segment}.t, blks{i_segment}.f(:,1), '.-');
    plot(ax2, blks{i_segment}.t, blks{i_segment}.a(:,1), '.-');
end
ylim(ax1, 330+20*[-1 1]);
xlabel(ax1, 'Time in attack [s]');
ylabel(ax1, 'Frequency f_1 [Hz]');
xlabel(ax2, 'Time in attack [s]');
ylabel(ax2, 'Growth rate \alpha_1 [1/s]');
leg = strcat({'Attack #'}, num2str((1:n_segments).'));
legend(ax1, leg, 'Location', 'NorthEast');