function [f0, alpha, err_fit] = estimate_fund_freq(x, fs, f0_appr)
%ESTIMATE_FUND_FREQ Estimate the fundamental frequency of a signal
%   [F0, ALPHA, ERR_FIT] = ESTIMATE_FUND_FREQ(X, FS)
%   [F0, ALPHA, ERR_FIT] = ESTIMATE_FUND_FREQ(X, FS, F0_APPR)
%
%   Input:
%       X - Signal (real valued column vector)
%       FS - Sampling freqeuncy (positive scalar)
%       F0_APPR - Approximate fundamental frequency (optional)
%
%   Output:
%       F0 - Estimated fundamental frequency (scalar)
%       ALPHA - Estimated temporal growth rate (scalar)
%       ERR_FIT - Relative L2 error of the least squares solution (scalar)

n_samp = size(x, 1);

if nargin < 3 || isempty(f0_appr)
    % No approximate frequency given, estimate using average spectrum
    n_spec = 4096;
    n_over = 2048;
    [X, f] = spec_aver(x, fs, n_spec, n_over, @hann);
    Y = hps(X, 3);
    Y(1) = 0;
    [~, idx] = max(Y);
    f0_appr = f(idx);
end

% Create the modulation signal
t_sect = (0 : n_samp-1).' / fs;
x_mod = exp(-2i*pi*f0_appr*t_sect);
    
f1_bw = 1/4;        % Minimal bandwidth compared to f_1
min_bw = 30;        % Absolute minimal bandwidth [Hz]

% Perform filtering and fitting
bw = max(f0_appr * f1_bw, min_bw);

% Create a butterworth filter
[b, a] = butter(5, bw / (fs/2));
y_filt = filter(b, a, bsxfun(@times, x, x_mod), [], 1);
    
% Drop the initial part of the evaluation (remove transient of filter)
eval_start_idx = find(t_sect > 10 / bw, 1, 'first');
    
% Assemble system of equations
lhs = y_filt(eval_start_idx : end - 1);
rhs = y_filt(eval_start_idx + 1 : end);
    
% Solve overdetermined system by least squares
gam = lhs \ rhs;
    
% Evaluate the error of the fit
err_fit = norm(gam * lhs - rhs) / norm(rhs);
    

df = angle(gam) / (2*pi) * fs;  % Frequency difference
f0 = f0_appr + df;              % Fundamental frequency
alpha = log(abs(gam)) * fs;     % Temporal growth rate

end % of function estimate_fund_freq

