function [blk, signal] = track_f_a(x, fs, f_n, opts)
%TRACK_F_A Track frequency and amplitude of partials
%   [BLK, SIGNAL] = TRACK_F_A(X, FS, F_N, OPTS)
%
%   Input:
%       X - input signal (column vector)
%       FS - sampling frequency (scalar)
%       F_N - guessed partial frequencies (vector)
%       OPTS - Options structure with fields [default values]
%           'block_size' - Block size in samples [256]
%           'block_overlap' - Block overlap in samples [128]
%           'bpf_order' - Band-pass filter order [3]
%           'lpf_order' - Low-pass filter order [5]
%           'min_bandwidth' - Minimum bandwidth of filter [FS*1e-3]
%           'min_frequency' - Minimum frequency to track [FS*1e-3]
%
%   Output:
%       BLK - structure holding block results with fields
%           t - time vector
%           A - amplitude matrix of harmonics
%           f - frequency matrix of harmonics
%           a - temporal growth rate of harmonics
%           gam_err - relative error of fitting gamma
%           fit_err - relative error of fitting
%       SIGNAL - filtered and fitted full signal structure with fields
%           filts - band-filtered signals
%           fitted - each harmonic fitted
%           fullfit - fitted full signal (sum of harmonics)

% Default options
default_opts = struct( ...
    'block_size', 256, ...          % Block size in samples
    'block_overlap', 128, ...       % Block overlap in samples
    'bpf_order', 3, ...             % Band-pass filter order
    'lpf_order', 5, ...             % Low-pass filter order
    'min_bandwidth', fs*1e-3, ...   % Minimal bandwidth for filter
    'min_frequency', fs*1e-3 ...    % Minimal frequency
);

% Fill the default options in the structure
if nargin < 4 || isempty(opts)
    opts = default_opts;
elseif ~isstruct(opts) || numel(opts) > 1
    error('track_f_a:invalid_arguments', ...
        'Input argument OPTS must be a structure')
else
    % Fill default values
    def_fields = fieldnames(default_opts).';
    fields = fieldnames(opts).';
    
    % Check if all fields are supported
    for fn = fields
        if ~ismember(fn, def_fields)
            warning('fill_default_options:unsupported_option', ...
                'Field %s of the structure OPTS is unknown', fn{1});
        end
    end
    
    % Copy default fields to options, where not found
    for fn = def_fields
        if ~isfield(opts, fn{1}) || isempty(opts.(fn{1}))
            opts.(fn{1}) = default_opts.(fn{1});
        end
    end
end
    
n_samp = size(x, 1);        % Number of samples
n_harm = numel(f_n);        % Number of harmonics (partials)

% Number of blocks
n_blocks = floor((n_samp - opts.block_size)/(opts.block_size - opts.block_overlap)) + 1;

% Preallocate for block quantitites
x_t = (0 : n_blocks-1) * (opts.block_size - opts.block_overlap)/fs;
x_f = zeros(n_blocks, n_harm);
x_a = zeros(n_blocks, n_harm);
x_amp = zeros(n_blocks, n_harm);
x_phi = zeros(n_blocks, n_harm);
x_err = zeros(n_blocks, n_harm);
x_ferr = zeros(n_blocks, n_harm);

% Preallocate for signal quantities
x_fitted = zeros(n_samp, n_harm);
x_full  = zeros(n_samp, 1);
x_filts = zeros(n_samp, n_harm);
    
% Create the low-pass filter
bw = max(min(f_n) * 1/3, opts.min_bandwidth);
[b, a] = butter(opts.lpf_order, bw/(fs/2));

% Create the time vector
t = (0 : n_samp-1).'/fs;

% Create the block window
if opts.block_size >= 2*opts.block_overlap
    blk_win = ones(opts.block_size, 1);
    blk_win(1 : opts.block_overlap) = (1 : opts.block_overlap) / opts.block_overlap;
    blk_win(opts.block_size - opts.block_overlap + (1 : opts.block_overlap)) = ...
        (opts.block_overlap-1 : -1 : 0) / opts.block_overlap;
else
    bs_ol = opts.block_size - opts.block_overlap;
    blk_win = zeros(opts.block_size, 1);
    blk_win(1 : bs_ol) = (1 : bs_ol) / bs_ol;
    blk_win(bs_ol + (1:bs_ol)) = (bs_ol-1 : -1 : 0) / bs_ol;
end

% Go through all partials
for i_harm = 1 : n_harm
    f_i = f_n(i_harm);
    f_lims = (f_i + bw * [-1 1]);
    f_lims(1) = max(f_lims(1), opts.min_frequency);
    % Create a bandpass filter near the peak
    [b_band, a_band] = butter(opts.bpf_order, f_lims  / (fs/2));
    % Filter the signal to the current band
    x_band = filtfilt(b_band, a_band, x);
    x_filts(:, i_harm) = x_band;

    % Filter without phase shift
    x_mod_filt = filtfilt(b, a, x_band .* exp(-2i*pi*f_i*t));
        
    % Time scale for a single block
    t_blk = (0 : opts.block_size-1).' / fs;
    
    % Go through all blocks
    for i_block = 1 : n_blocks
        % Indices in the current block
        idxs = (i_block-1)*(opts.block_size - opts.block_overlap) + (1 : opts.block_size).';
        
        % Calculate the exponential fit
        lhs = x_mod_filt(idxs(1:end-1));
        rhs = x_mod_filt(idxs(2:end));
        
        gam = lhs \ rhs;
        
        % Calculate the error of the exponential fit
        x_ferr(i_block, i_harm) = norm(gam * lhs - rhs) / norm(rhs);
        
        amp = log(abs(gam)) * fs;           % Amplification (exp(amp*t))
        df = angle(gam) / (2*pi) * fs;      % Frequency deviation
        
        % Sine and cosine components
        s = exp(amp*t_blk) .* sin(2*pi*(f_i + df)*t_blk);
        c = exp(amp*t_blk) .* cos(2*pi*(f_i + df)*t_blk);
        
        % Fit to the band-limited signal
        w = [s, c] \ x_band(idxs);
        
        % Block amplitude at the middle of the block
        x_amp(i_block, i_harm) = sqrt(w(1).^2 + w(2).^2) ... 
            * exp(amp*t_blk(floor(opts.block_size/2)));
        
        % Block phase shift
        x_phi(i_block, i_harm) = atan2(w(2), w(1));
        
        % Fitted attack transient
        blk_fit = [s, c] * w;
       
        x_fitted(idxs, i_harm) = x_fitted(idxs, i_harm) + blk_fit .* blk_win;
        x_full(idxs) = x_full(idxs) + blk_fit .* blk_win;
        
        % Calculate error
        x_err(i_block, i_harm) = ...
            norm(x_fitted(idxs, i_harm)-x_band(idxs)) / norm(x_band(idxs));
        
        x_a(i_block, i_harm) = amp;
        x_f(i_block, i_harm) = f_i + df;
    end % of cycle over blocks
end % of cycle over harmonics

% Assign output variables
blk = struct( ...
    't', x_t, ...
    'A', x_amp, ...
    'f', x_f, ...
    'a', x_a, ...
    'phi', x_phi, ...
    'fit_err', x_err, ...
    'gam_err', x_ferr ...
);

signal = struct( ...
    'filts', x_filts, ...
    'fitted', x_fitted, ...
    'fullfit', x_full ...
);

end % of track_f_a

